/*

 ReaderResultBuilder.java
 
 Copyright 2004 KUBO Hiroya (hiroya@sfc.keio.ac.jp).
 
 Licensed under the Apache License, Version 2.0 (the "License");
 you may not use this file except in compliance with the License.
 You may obtain a copy of the License at
 
 http://www.apache.org/licenses/LICENSE-2.0
 
 Unless required by applicable law or agreed to in writing, software
 distributed under the License is distributed on an "AS IS" BASIS,
 WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 See the License for the specific language governing permissions and
 limitations under the License.
 
 Created on 2004/11/01

 */
package net.sf.sqs_xml.reader.logic;


import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.NumberFormat;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import javax.xml.transform.TransformerException;

import net.sf.sqs_xml.util.FileUtil;
import net.sf.sqs_xml.util.StringUtil;
import net.sf.sqs_xml.xmlns.SQSNamespaces;

/**
 * @author hiroya
 *
 */
public class ReaderResultBuilder {

    public static final int ALL_MODE = 0;
    public static final int TEXTAREA_MODE = 1;
    public static final int MARK_MODE = 2;
    protected Object[] matrix = null;
    protected List rowFileList = new LinkedList();

    private Map errorImageMap = new HashMap();
    private Map areaImageMap = new HashMap();
    Map cellIDToAreaFileMap = new HashMap();
    
    protected QuestionnaireSchema schema;
    private List currentRowFileList;
    
    public ReaderResultBuilder(){
    }
    
    public void setQuestionnaireSchema(QuestionnaireSchema schema){
        this.schema = schema;        
    }
    
    public void setRowSize(int size){
        this.matrix = new Object[size];
    }
   
    public void addRowFileList(List rowFileList){
        this.rowFileList.add(rowFileList);
        currentRowFileList = rowFileList;
    }

    public void addRowFile(File rowFile){
        this.currentRowFileList.add(rowFile);
    }
    
    public void setMatrix(int index, Object[] line){
        this.matrix[index] = line;
    }

    public List getColumnNames(){
        return this.schema.columnNames;
    }
    
    public List getColumnLabels(){
        return this.schema.columnLabels;
    }
    
    public int getPage(String id){
        try{
            return schema.getPage(id);
        }catch(TransformerException ex){
            return -1;
        }
    }
    
    class CellID{
        int rowId;
        String id;
        CellID(int rowId, String id){
            this.rowId = rowId;
            this.id = id;
        }
        public boolean equals(Object o){
            if(o == null && !(o instanceof CellID)){
                return false;
            }else{
                CellID c = (CellID)o;
                return c.rowId == rowId && c.id.equals(id); 
            }
        }
        public int hashCode(){
            return rowId+id.hashCode();
        }
    }
    
    public void putAreaFile(int rowId, String id, File areaFile){
        cellIDToAreaFileMap.put(new CellID(rowId, id), areaFile);
    }
    
    public File getAreaFile(int rowId, String id){
        return (File)cellIDToAreaFileMap.get(new CellID(rowId, id));
    }

    public void addAreaImage(int rowIndex, String qid, File file){
        areaImageMap.put(rowIndex+"\t"+qid, file);
    }
    public void addErrorImage(int rowIndex, String qid, File file){
        errorImageMap.put(rowIndex+"\t"+qid, file);
    }

    public void saveCSV(File tgtfile, String encoding) throws TransformerException, IOException {
        PrintWriter out = FileUtil.createPrintWriter(tgtfile, encoding);
         saveCSVHeader(out);
         saveCVSBody(out);
         out.close();
     }

    public void saveHTML(File tgtfile, String encoding, File imagedir, File resultdir, int mode) throws TransformerException, IOException {
        PrintWriter out = FileUtil.createPrintWriter(tgtfile, encoding);
        saveHTML(out, encoding, imagedir, resultdir, tgtfile, mode);
        out.close();
     }

    /**
     * @param out
     */
    private void saveCVSBody(PrintWriter out) {
        for(int j=0; j < matrix.length; j++){
            out.print(j+1);
            out.print("\t");
            out.print(((List)rowFileList.get(j)).get(0));
            Object[] line = (Object[])matrix[j];
            saveCSVRow(out, j, line);
            out.println();
        }
    }

    private void saveCSVRow(PrintWriter out, int rowID, Object[] line) {
        for(int columnID=0; columnID < line.length; columnID++){
            out.print("\t");
            List columnNameList = getColumnNames();
            if(line[columnID] != null){
                String value = null;
                if(line[columnID] instanceof String){
                    value = (String)line[columnID];
                }else if(line[columnID] instanceof CellValueOption){
                    value = ((CellValueOption)line[columnID]).value;                        
                }
                String id = (String)columnNameList.get(columnID);
                File errorImageFile = (File)errorImageMap.get(rowID+"\t"+id);
                if(errorImageFile != null){// && areaImageFile.getName().endsWith("-MarkError.png")
                    out.print("?");
                }else{
                    out.print(StringUtil.escape(value));
                }
            }
        }
    }

    /**
     * @param out
     */
    private void saveHTML(PrintWriter out, String encoding, File imagedir, File resultdir, File targetFile, int mode) throws TransformerException {
        try{
        out.println("<?xml version=\"1.0\" encoding=\""+encoding+"\"?>");
        //out.println("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.1//EN\" \"http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd\">");
        out.println("<html xmlns=\""+SQSNamespaces.XHTML_URI+"\" xml:lang=\"ja\" lang=\"ja\" version=\"1.1\">");
        out.println("<head>");
        out.println("<meta http-equiv=\"Content-Type\" content=\"text/html; charset="+encoding+"\" />");
        out.print("<title>");
        out.print(targetFile.getAbsolutePath());
        out.println("</title>");
        out.println("<style type=\"text/css\">table, th, td{border-style: solid; border-width: 1px; border-color: black;} tr.page{background-color:#bbf}tr.qid{background-color:#fbb}tr.foot{background-color:#fbb} tr.foot2{background-color:#ffb} th.id{background-color:#bfb} img{border-width:0} body{background-color: #ddd; padding: 0.5em} textarea{background-color:#fff}</style></head>");
        out.println("<body>");
        out.println("<h1>"+targetFile.getAbsolutePath()+"</h1>");
        out.print("<table>");
        out.println("<thead>");
        out.print("<tr class=\"page\"><th rowspan=\"2\">sampleID</th>");
        out.print("<th rowspan=\"2\">file</th>");
        out.println();
        saveHTMLTableHeader1(out, imagedir, mode);
        out.print("</tr>");
        out.print("<tr class=\"qid\">");
        saveHTMLTableHeader2(out, mode);
        out.print("</tr>");
        out.println("</thead><tbody>");
        saveHTMLRows(out, imagedir, mode);
        out.print("</tbody>");
        out.print("<tfoot>");
        saveHTMLTableFooter(out, mode);
        out.print("</tfoot>");
        out.println("</table>");
        out.println("</body></html>");
        }catch(Exception ex){
            ex.printStackTrace();
        }
    }

    private void saveHTMLRows(PrintWriter out, File imagedir, int mode) throws TransformerException{
        for(int rowID=0; rowID < matrix.length; rowID++){
            out.print("<tr>");
            out.print("<th class=\"id\">");
            out.print(rowID+1);
            out.print("</th>");
            out.print("<th class=\"id\">");//<ol>
            List rowFiles = (List)rowFileList.get(rowID);
            for(int i = 0; i < rowFiles.size(); i++){
                if(0 < i){
                    out.print(", ");
                }
                File file = ((File)rowFiles.get(i));//<li>
                out.print("<a href=\"../"+FileFunctions.getRelativePath(imagedir, file)+"\" title=\""+file.getAbsolutePath()+"\">");
                //out.print(file.getName());
                out.print(i+1);
                out.print("</a>");//</li>
            }
            out.print("</th>");//</ol>
            saveHTMLRowColumns(out, imagedir, mode, rowID);
            out.print("</tr>");
            out.println();
        }
    }

    private void saveHTMLTableHeader1(PrintWriter out, File imagedir, int mode) throws TransformerException {
        int page, prevPage = -1;
        int colspan = 0;
        for(int columnID = 0; columnID < getColumnNames().size(); columnID++){
            String id = (String)getColumnNames().get(columnID);          
            if((mode == TEXTAREA_MODE) && getAreaFile(0, id) == null){
                continue;
            }
            String columnName = (String)getColumnNames().get(columnID);
            page = getPage(columnName);
            String style = createStyle(columnID, id);
            out.print("<th style=\""+style+"\">");
            File file = (File)((List)rowFileList.get(0)).get(page-1);
            out.print("<a href=\"../"+FileFunctions.getRelativePath(imagedir, file)+"\">P."+page+"</a>");//tabindex=\""+columnID+"\" 
            out.println("</th>");
        }
    }

    private void saveHTMLTableHeader2(PrintWriter out, int mode) throws TransformerException {
        for(int columnID = 0; columnID < getColumnNames().size(); columnID++){
            String id = (String)getColumnNames().get(columnID);          
            if((mode == TEXTAREA_MODE) && getAreaFile(0, id) == null){
                continue;
            }
            String columnName = (String)getColumnNames().get(columnID);
            String style = createStyle(columnID, id);
            out.print("<th style=\""+style+"\">");
            out.print(StringUtil.replaceAll(columnName, "/", "/<br/>"));
            //out.println(getColumnLabels().get(columnID));
            out.println("</th>");
        }
    }

    private void saveHTMLTableFooter(PrintWriter out, int mode) throws TransformerException {
        int len = getColumnNames().size();
        int[] totals = new int[len] ;
        int[] valids = new int[len] ;
        
        if(mode == TEXTAREA_MODE){
            return;
         }
                
        for(int columnID = 0; columnID < len; columnID++){
            String id = (String)getColumnNames().get(columnID);          
            String columnName = (String)getColumnNames().get(columnID);
            for(int rowID = 0; rowID < matrix.length; rowID++){
                Object[] line = (Object[])matrix[rowID];
                try{
                    String value = null;
                    if(line[columnID] == null){
                        continue;
                 }else if(line[columnID] instanceof CellValueOption){
                     value = ((CellValueOption)line[columnID]).value;
                 }else{
                     value = (String)line[columnID];
                     }
                    totals[columnID] += Integer.parseInt(value);
                    valids[columnID]++;
               }catch(Exception ignore){
                   //ignore.printStackTrace();
                  }
              }            
        }
        
        NumberFormat format = NumberFormat.getInstance();
        format.setMaximumFractionDigits(3);
        for(int i = 0; i < 6; i++){
            if(i == 0){
                out.print("<tr style=\"background-color: #f8f\"><th colspan=\"2\">");
            }else if(1 <= i && i <= 3){                
                out.print("<tr class=\"foot\"><th colspan=\"2\">");
            }else{
                out.print("<tr class=\"foot2\"><th colspan=\"2\">");
              }
            if(i == 0){
                out.print("合計値"); 
            }else if(i == 1){
                out.print("有効回答数");
            }else if(i == 2){
                out.print("平均値"); 
            }else if(i == 3){
                out.print("標準偏差"); 
            }else if(i == 4){
                out.print("平均値(全体)"); 
            }else if(i == 5){
                out.print("標準偏差(全体)"); 
              }
            out.print("</th>");
            
            for(int columnID = 0; columnID < len; columnID++){
                String id = (String)getColumnNames().get(columnID);
                if((mode == TEXTAREA_MODE) && getAreaFile(0, id) == null){
                    continue;
                   }
                
                if(valids[columnID] == 0){
                        continue;
                   }
                
                String style = createStyle(columnID, id);
                out.print("<td style=\""+style+"\">");
                if( i == 0){
                    out.print(totals[columnID]);                    
                } else if(i == 1){
                    out.print(valids[columnID]);
                } else if(i == 2){
                    if(1 <= valids[columnID]){
                        double average = 1.0 * totals[columnID] / valids[columnID] ;
                        out.print(format.format(average));
                        }
                } else if(i == 3){
                    if(0<valids[columnID]){
                        double average = 1.0 * totals[columnID] / valids[columnID] ;
                        double diff = calcDiff(columnID, average);
                        out.print(format.format(Math.sqrt(diff/valids[columnID])));
                       }
                } else if(i == 4){
                    if(1 <= matrix.length){
                        double average = 1.0 * totals[columnID] / matrix.length;
                        out.print(format.format(average));
                        }
                } else if(i == 5){
                    if(0 < matrix.length){
                        double average = 1.0 * totals[columnID] / matrix.length ;
                        double diff = calcDiff(columnID, average);
                        out.print(format.format(Math.sqrt(diff / matrix.length)));
                       }
                   }
                out.println("</td>");
            }
            out.print("</tr>");
        }
    }

    private double calcDiff(int columnID, double average) {
        double diff = 0.0;
        for(int rowID = 0; rowID < matrix.length; rowID++){
            Object[] line = (Object[])matrix[rowID];
            String value = null;
            if(line[columnID] == null){
                continue;
            }else if(line[columnID] instanceof CellValueOption){
                value = ((CellValueOption)line[columnID]).value;
            }else{
                value = (String)line[columnID];
                 }
              try{
               double d = (average - Integer.parseInt(value));
               diff += d * d;
              }catch(NumberFormatException ignore){
                 }
             }
        return diff;
    }
    
    boolean isPageMargin(int columnID, String id)throws TransformerException{
        int page = schema.getPage(id);
        if(1 < columnID && page != schema.getPage((String)getColumnNames().get(columnID - 1))){
            return true;
         }
        return false;
    }
    
    boolean isColumnMargin(int columnID, String id){
        if(columnID < 1){
            return false;
         }
    
        String prevID = (String)getColumnNames().get(columnID - 1);
        int vp = id.lastIndexOf("/");
        int prevVp = prevID.lastIndexOf("/");
        if(vp < 0 && prevVp < 0){
            return false;
        }
        
        if((vp < 0 && 0 <= prevVp) || (prevVp < 0 && 0 <= vp) ||
                ! id.substring(0, vp).equals(prevID.substring(0, prevVp))){
            return true;
        }
        return false;
    }
        
        
        /*
        String borderWidth = (page == prevPage)? "":"border-left-style: double; border-left-width: 4px;";
        prevPage = page;
        String borderColor = "";
        int vp = id.lastIndexOf("/v");
        String qid = id; 
        if(0 < vp && ! (qid = id.substring(0, vp)).equals(prevQID)){
            borderColor = "";
        }else{
            int sp = id.lastIndexOf("/");
            if(0 < sp && ! (qid = id.substring(0, sp)).equals(prevQID)){
                borderColor = "";
                borderWidth = "border-left-width: 2px;";
                }
          }
        prevQID = qid;
        return true;
    }*/

    private void saveHTMLRowColumns(PrintWriter out, File imagedir, int mode, int rowID) throws TransformerException {
        Object[] line = (Object[])matrix[rowID];
        for(int columnID = 0; columnID < line.length; columnID++){
            String id = (String)getColumnNames().get(columnID);
            File imagefile = getAreaFile(rowID, id);
            
            if((mode == TEXTAREA_MODE) && imagefile == null){
                continue;
            }
            
            String style = createStyle(columnID, id);
            
            File errorImageFile = (File)errorImageMap.get(rowID+"\t"+id);
            if(errorImageFile != null){// && areaImageFile.getName().endsWith("-MarkError.png")
                out.print("<td style=\"background-color: #faf; "+style+"\">");
                out.print("<a href=\"ERROR/"+errorImageFile.getName()+"\">?</a>");
                out.print("</td>");
                continue;
            }
            String bgcolor = (rowID % 2 == 0)? "":"background-color: #ccf;";
            out.print("<td style=\""+bgcolor+style+"\">");
            
            String value = null;
            if(line[columnID] != null){
                if(line[columnID] instanceof String){
                    value = (String)line[columnID];
                }else if(line[columnID] instanceof CellValueOption){
                    value = ((CellValueOption)line[columnID]).value;
                }else{
                    value = line[columnID].getClass().getName();
                   }
              }
            
            File areaImageFile = (File)areaImageMap.get(rowID+"\t"+id);
            if((mode == TEXTAREA_MODE || mode == ALL_MODE) && imagefile != null){
                out.print("<a href=\"../"+FileFunctions.getRelativePath(imagedir, areaImageFile)+"\" title=\""+(rowID+1)+"-"+id+"\" tabindex=\""+columnID+"\">");
                out.print("<img src=\"TEXTAREA/"+imagefile.getName()+"\" alt=\""+(rowID+1)+"-"+id+"\"/>");
                out.print("</a>");
            }
                
            if(value != null){
                out.print(StringUtil.escapeHTML(value));
            }else{
                out.print(" ");
            }
          out.print("</td>");
        }
    }

    private String createStyle(int columnID, String id) throws TransformerException {
        String style = "";
        if(isPageMargin(columnID, id)){
            style = "border-left-style: solid;border-left-color: black; border-left-width: 8px;";
        }else if(isColumnMargin(columnID, id)){
            style = "border-left-style: dotted; border-left-color: red;border-left-width: 3px;";
          }
        return style;
    }

    /**
     * @param out
     * @param columnNames
     */
    private void saveCSVHeader(PrintWriter out) throws TransformerException {
        List columnNames = getColumnNames();
        out.print("sampleID");
        out.print("\t");
        out.print("file");
        for(int i=0; i < columnNames.size(); i++){
            out.print("\t");
            String id = (String)columnNames.get(i);
            out.print(id);
        }
        out.println();
    }
}
